program arctan2_speed;

uses
  sysutils, Math,crt;

const
  MILLION = 1000 * 1000;
  N = 20 * MILLION;

type
  TArcTan2Func = function(y, x: Float): Float;

  TData = record
    x, y: Float;
  end;

var
  Data: array[1..n] of TData;
  j,m,p:Array[1..n] of Float;
  ox,oy:byte;
  mt,pt,jt,bt:double;
  mc,pc,jc:longint;
  tc:boolean=false;
  firstrun:boolean=true;

{$define modedmatharctan}

function matharctan2(y,x : float) : float;
begin
  if (x=0) then
      begin
        if y=0 then
          matharctan2:=0.0
        else if y>0 then
          matharctan2:=pi/2
        else if y<0 then
          matharctan2:=-pi/2;
      end
  else
{$ifdef modedmatharctan}
    if x<0.0 then
      begin
        if y<0.0 then matharctan2:=arctan(y/x)-pi
        else matharctan2:= arctan(y/x)+pi;
      end
      else matharctan2:=ArcTan(y/x) ;
 {$else}
    matharctan2:=ArcTan(y/x);
    if x<0.0 then
      matharctan2:=matharctan2+pi;
    if matharctan2>pi then
      matharctan2:=matharctan2-2*pi;
 {$endif}
end;

function PaoloArcTan2(Y, X : Float) : Float;
begin
  if (X = 0) then begin
    if (Y = 0) then
      Result:=0
    else
      if (Y > 0) then
        Result:=0.5*pi
      else
        Result:=-0.5*pi
  end
  else begin
    Result:=ArcTan(Y/X);
    if (X < 0) then
      if (Y < 0) then
        Result:=Result-pi
      else
        Result:=Result+pi
  end;
end;

function JoshArcTan2 (y, x : float) : Float;
begin
  if x>0.0 then result:=arctan(y/x)
  else
    if x<0.0 then if y<0.0 then result:=arctan(y/x)-pi else result:= arctan(y/x)+pi
    else {x must be zero} if y<0.0 then result:=-pi/2.0 else if y>0.0 then result:=pi/2.0
                                            else result:=0.0; // x=0 y=0 so set to 0
end;



  
procedure PrepareData(fixx,fixy:boolean;v:float);
var
  i: Integer;
begin
 // randomize;
 // RandSeed := 0;
  for i := 1 to N do
  begin
    if fixx then Data[i].X:=v else Data[i].X :=(Random()-0.5) * high(dword);
    if fixy then Data[i].Y:=v else Data[i].Y :=(Random()-0.5) * high(dword);
  end;
end;

procedure displaytimes;
var dx:integer=48;
    dy:integer=1;
begin
  cursoroff;
  TextBackground(Blue);
  oy:=wherey;ox:=wherex;
  gotoxy(dx,dy);textcolor(LightGray);write('        Total Times   Count '+mc.ToString); clreol;
  gotoxy(dx,dy+1);textcolor(LightGray);write('        hr mn se ms     Avg');  clreol;
  textcolor(yellow);
  if mc>0 then
  begin
    gotoxy(dx,dy+2);
    write(' Math  :', FormatDateTime('hh.nn.ss.zzz', mt), '  ',FormatDateTime('s.zzz', mt/mc), ' sec'); clreol;
  end;
  if pc>0 then
  begin
    gotoxy(dx,dy+3);   if pc>mc then textcolor(red);
    write(' Paolo :', FormatDateTime('hh.nn.ss.zzz', pt), '  ',FormatDateTime('s.zzz', pt/pc), ' sec');  clreol;
  end;
  if jc>0 then
  begin
    gotoxy(dx,dy+4);  if jc>mc then textcolor(red);
    write(' Josh  :', FormatDateTime('hh.nn.ss.zzz', jt), '  ',FormatDateTime('s.zzz', jt/jc), ' sec');  clreol;
  end;
 // dx:
//  gotoxy(dx,dy+5);write(':-------------------------------');  clreol;
  gotoxy(ox,oy);
  if tc then textcolor(white) else textcolor(DarkGray);
  TextBackground(black);
end;

procedure Test(Descr: String; ArcTan2Func: TArcTan2Func; w:string);
var
  t: TDateTime;
  i: Integer;
  ans:float;
begin
  Write('  Testing ' + Descr + '...');
  RandSeed := 0;
  t := Now;
  for i := 1 to N do
  begin
    ans:=ArcTan2Func(Data[i].Y, Data[i].X);
    if w='j' then j[i]:=ans;
    if w='m' then m[i]:=ans;
    if w='p' then p[i]:=ans;
  end;
  t := Now - t;
  if w='m' then bt:=t;
  if t>bt then textcolor(red);
  WriteLn(' done: ', FormatDateTime('s.zzz', t), ' sec');
  if tc then textcolor(white) else textcolor(DarkGray);
  if w='j' then
  begin
    jt:=jt+t;
    inc(jc);
  end;
  if w='m' then
  begin
    mt:=mt+t;
    inc(mc);
  end;
  if w='p' then
  begin
    pt:=pt+t;
    inc(pc);
  end;
//  displaytimes;
end;
var l,c:integer;
  td:extended;

procedure reportresult(heading:string);
begin
  tc:=not tc;
  if tc then textcolor(white) else textcolor(DarkGray);
  WriteLn(heading);
  Test({$ifdef modedmatharctan}'*Matharctan2'{$else}'Matharctan2 '{$endif}, @matharctan2,'m');
  Test('PaoloArctan2', @paoloarctan2,'p');
  Test('JoshArctan2 ', @josharctan2,'j');
 // WriteLn('Finished. Calculating');
//  Write('  Check diff between math and Josh');
  c:=0;
  td:=0;
  for l:=1 to n do
  begin
    if j[l]<>m[l] then
      begin
        inc(c);
        td:=td+abs(j[l]-m[l]);
      end;
  end;
  if c=0 then WriteLn('Math and Josh SAME')
  else
  begin
  //  WriteLn;
    WriteLn('Josh Differs  '+c.ToString+' Total of Differences ',td:4:16);
  end;
//  Write('  Check diff between math and Paolo');
  c:=0;
  td:=0;
  for l:=1 to n do
  begin
    if m[l]<>p[l] then
      begin
        inc(c);
        td:=td+abs(p[l]-m[l]);
      end;
  end;
  if c=0 then WriteLn('Math and Paolo SAME')
  else
  begin
  //  WriteLn;
    WriteLn('Paolo Differs  '+c.ToString+' Total of Differences ',td:4:16);
  end;
 // Writeln('** End of Test ** '+heading);
//  Writeln;//Writeln;
end;

begin
  mt:=0;pt:=0;jt:=0;
  pc:=0;pc:=0;jc:=0;
  repeat
  randomize;
  RandSeed := 0;
  if firstrun then PrepareData(false,false,0);
  firstrun:=false;
  PrepareData(false,false,0);
  if keypressed then exit;
  gotoxy(1,1);
  reportresult('Calculating and testing 20M random x/y samples');
  PrepareData(true,false,0);
  displaytimes;
  if keypressed then exit;
  reportresult('Calculating and testing 20M samples with all x=0');
  PrepareData(false,true,0);
  displaytimes;
  if keypressed then exit;
  reportresult('Calculating and testing 20M samples with all y=0');
  PrepareData(true,true,0);
  displaytimes;
  if keypressed then exit;
  reportresult('Calculating and testing 20M samples with all y=0 and x=0');
  if keypressed then exit;
  //delay(5000);
  displaytimes;
  tc:=not tc;
  until keypressed;
end.

